/*
 * DISK.ASM
 * ========
 *
 * This is what the disk image is created from. There several
 * segments contained in this file.
 *
 * First is the BIOS Parameter Block (BPB), which is part of the FAT
 * specification. It lays out various values needed for the FAT
 * filesystem.
 *
 * Second that is the initialization of the File Allocation Table.
 * This only needs a couple of entries to be initialized.
 *
 * Third is the Root Directory Volume ID entry. This needs to be
 * created on initialization.
 *
 * Fourth is the Game Boy Cartridge Header. This contains various
 * values, keys, and checksums to make the image a valid ROM file.
 *
 * Fifth is inturrupt handler routines. This bootloader sets up
 * interrupt vectors in HRAM that can be changed during runtime.
 *
 * Sixth is the bootloader itself. It is broken up into sections of
 * code to try and squeeze it into all the little gaps between other
 * portions of the boot sector. The bootloader functions by locating
 * the file "\KERNEL.BIN" in the filesystem, copying it into WRAM
 * at $C000, and jumping to $C000.
 *
 * Seventh is the VHD Footer, which resides at the very end of the
 * image. It defines the image as a Virtual Hard Drive image,
 * hopefully making it easier to mount the image under MS Windows.
 */

include "CONSTANT.INC"

/*
 * FAT specification found at https://academy.cba.mit.edu/classes/networking_communications/SD/FAT.pdf
 */

section "BS_jmpBoot", rom0[$0000]
/*
 * Jump instruction to boot code. This field has two
 * allowed forms:
 * jmpBoot[0] = 0xEB, jmpBoot[1] = 0x??,
 * jmpBoot[2] = 0x90
 * and
 * jmpBoot[0] = 0xE9, jmpBoot[1] = 0x??,
 * jmpBoot[2] = 0x??
 * 0x?? indicates that any 8-bit value is allowed in
 * that byte. What this forms is a three-byte Intel x86
 * unconditional branch (jump) instruction that jumps
 * to the start of the operating system bootstrap code.
 * This code typically occupies the rest of sector 0 of
 * the volume following the BPB and possibly other
 * sectors. Either of these forms is acceptable.
 * JmpBoot[0] = 0xEB is the more frequently used
 * format.
 *
 * Size: 3 bytes
 */
BS_jmpBoot:
    db $E9
    dw x86_err - @ - 2

section "x86_err", rom0
; This section is x86 machine code.
x86_err:
    db $CD, $18; INT 18h: diskless boot hook (start basic)
.loop:
    db $EB, .loop - @; JMP .loop

section "BS_OEMName", rom0[$0003]
/*
 * OEM Name Identifier. Can be set by a FAT
 * implementation to any desired value.
 * Typically this is some indication of what system
 * formatted the volume.
 *
 * Size: 8 bytes
 */
BS_OEMName:
    db "{GBDOS_NAME_STRING}"
    ds $000B - @, 0

section "BPB_BytsPerSec", rom0[$000B]
/*
 * Count of bytes per sector. This value may take on
 * only the following values: 512, 1024, 2048 or 4096.
 *
 * Size: 2 bytes
 */
BPB_BytsPerSec:
    dw BYTES_PER_SECTOR

section "BPB_SecPerClus", rom0[$000D]
/*
 * Number of sectors per allocation unit. This value
 * must be a power of 2 that is greater than 0. The
 * legal values are 1, 2, 4, 8, 16, 32, 64, and 128.
 *
 * Size: 1 byte
 */
BPB_SecPerClus:
    db SECTORS_PER_CLUSTER

section "BPB_RsvdSecCnt", rom0[$000E]
/*
 * Number of reserved sectors in the reserved region
 * of the volume starting at the first sector of the
 * volume. This field is used to align the start of the
 * data area to integral multiples of the cluster size
 * with respect to the start of the partition/media.
 * This field must not be 0 and can be any non-zero
 * value.
 * This field should typically be used to align the start
 * of the data area (cluster #2) to the desired
 * alignment unit, typically cluster size.
 *
 * Size: 2 bytes
 */
BPB_RsvdSecCnt:
    dw RESERVED_SECTOR_COUNT

section "BPB_NumFATs", rom0[$0010]
/*
 * The count of file allocation tables (FATs) on the
 * volume. A value of 2 is recommended although a
 * value of 1 is acceptable.
 *
 * Size: 1 byte
 */
BPB_NumFATs:
    db NUMBER_OF_FATS

section "BPB_RootEntCnt", rom0[$0011]
/*
 * For FAT12 and FAT16 volumes, this field contains
 * the count of 32-byte directory entries in the root
 * directory. For FAT32 volumes, this field must be set
 * to 0. For FAT12 and FAT16 volumes, this value
 * should always specify a count that when multiplied
 * by 32 results in an even multiple of
 * BPB_BytsPerSec.
 * For maximum compatibility, FAT16 volumes should
 * use the value 512.
 *
 * Size: 2 bytes
 */
BPB_RootEntCnt:
    dw ROOT_ENTRY_COUNT

section "BPB_TotSec16", rom0[$0013]
/*
 * This field is the old 16-bit total count of sectors on
 * the volume. This count includes the count of all
 * sectors in all four regions of the volume.
 * This field can be 0; if it is 0, then BPB_TotSec32
 * nmust be non-zero. For FAT32 volumes, this field
 * must be 0.
 * For FAT12 and FAT16 volumes, this field contains
 * the sector count, and BPB_TotSec32 is 0 if the
 * total sector count “fits” (is less than 0x10000).
 *
 * Size: 2 bytes
 */
BPB_TotSec16:
    dw TOTAL_SECTOR_COUNT

section "BPB_Media", rom0[$0015]
/*
 * The legal values for this field are 0xF0, 0xF8, 0xF9,
 * 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, and 0xFF.
 * 0xF8 is the standard value for “fixed” (nonremovable) media. For
 * removable media, 0xF0 is
 * frequently used.
 *
 * Size: 1 byte
 */
BPB_Media:
    db $F8

section "BPB_FATSz16", rom0[$0016]
/*
 * This field is the FAT12/FAT16 16-bit count of
 * sectors occupied by one FAT. On FAT32 volumes
 * this field must be 0, and BPB_FATSz32 contains
 * the FAT size count.
 *
 * Size: 2 bytes
 */
BPB_FatSz16:
    dw FAT_SIZE

section "BPB_SecPerTrk", rom0[$0018]
/*
 * Sectors per track for interrupt 0x13.
 * This field is only relevant for media that have a
 * geometry (volume is broken down into tracks by
 * multiple heads and cylinders) and are visible on
 * interrupt 0x13.
 *
 * Size: 2 bytes
 */
BPB_SecPerTrk:
    dw $4000 / BYTES_PER_SECTOR; Interpreted as sectors per bank

section "BPB_NumHeads", rom0[$001A]
/*
 * Number of heads for interrupt 0x13. This field is
 * relevant as discussed earlier for BPB_SecPerTrk.
 * This field contains the one based “count of heads”.
 * For example, on a 1.44 MB 3.5-inch floppy drive
 * this value is 2.
 *
 * Size: 2 bytes
 */
BPB_NumHeads:
    dw 1

section "BPB_HiddSec", rom0[$001C]
/*
 * Count of hidden sectors preceding the partition that
 * contains this FAT volume. This field is generally
 * only relevant for media visible on interrupt 0x13.
 * This field must always be zero on media that are
 * not partitioned.
 * NOTE: Attempting to utilize this field to align the
 * start of data area is incorrect
 *
 * Size: 4 bytes
 */
BPB_HiddSec:
    dl 0

section "BPB_TotSec32", rom0[$0020]
/*
 * This field is the new 32-bit total count of sectors on
 * the volume. This count includes the count of all
 * sectors in all four regions of the volume.
 * This field can be 0; if it is 0, then BPB_TotSec16
 * must be non-zero. For FAT12/FAT16 volumes, this
 * field contains the sector count if BPB_TotSec16 is
 * 0 (count is greater than or equal to 0x10000).
 * For FAT32 volumes, this field must be non-zero.
 *
 * Size: 4 bytes
 */
BPB_TotSec32:
    dl 0

section "BS_DrvNum", rom0[$0024]
/*
 * Interrupt 0x13 drive number. Set value to 0x80 or
 * 0x00.
 *
 * Size: 1 byte
 */
BS_DrvNum:
    db $80

section "BS_Reserved1", rom0[$0025]
/*
 * Reserved. Set value to 0x0.0
 *
 * Size: 1 byte
 */
BS_Reserved1:
    db 0

section "BS_BootSig", rom0[$0026]
/*
 * Extended boot signature. Set value to 0x29 if
 * either of the following two fields are non-zero.
 * This is a signature byte that indicates that the
 * following three fields in the boot sector are
 * present.
 *
 * Size: 1 byte
 */
 BS_BootSig:
    db $29

section "BS_VolID", rom0[$0027]
/*
 * Volume serial number.
 * This field, together with BS_VolLab, supports
 * volume tracking on removable media. These
 * values allow FAT file system drivers to detect that
 * the wrong disk is inserted in a removable drive.
 * This ID should be generated by simply combining
 * the current date and time into a 32-bit value.
 *
 * Size: 4 bytes
 */
; 76543210 76543210 76543210 76543210
; [Yea.][M .][Da.][ Hr.][Min .][Sec.]
BS_VolID:
    dl ( ((__UTC_YEAR__ % %11111) << 26) \
       + (__UTC_MONTH__ << 22) \
       + (__UTC_DAY__ << 17) \
       + (__UTC_HOUR__ << 12) \
       + (__UTC_MINUTE__ << 6) \
       + (__UTC_SECOND__) \
       )

section "BS_VolLab", rom0[$002B]
/*
 * Volume label. This field matches the 11-byte
 * volume label recorded in the root directory.
 * NOTE: FAT file system drivers must ensure that
 * they update this field when the volume label file in
 * the root directory has its name changed or
 * created. The setting for this field when there is no
 * volume label is the string “NO NAME    ”.
 *
 * Size: 11 bytes
 */
BS_VolLab:
    db "{VOLUME_LABEL}"

section "BS_FilSysType", rom0[$0036]
/*
 * One of the strings “FAT12 ”, “FAT16 ”, or “FAT
 * ”.
 * NOTE: This string is informational only and does
 * not determine the FAT type.
 *
 * Size: 8 bytes
 */
BS_FilSysType:
if FAT_TYPE == 12
    db "FAT12"
else
    db "FAT16"
endc
    ds $003E - @, " "


section "Signature_word", rom0[$01FE]
/*
 * Set to 0x55 (at byte offset 510) and 0xAA (at byte
 * offset 511)
 */
Signature_word:
    db $55, $AA

/*
 * Initialize the FAT
 */
section "Reserved_FAT_entries", rom0[RESERVED_SECTOR_COUNT * BYTES_PER_SECTOR]
if FAT_TYPE == 12
    db $F8, $FF, $FF
else ; FAT16
    dw $FFF8, $FFFF
endc

/*
 * Root Directory Volume ID Entry
 */
if (RESERVED_SECTOR_COUNT + FAT_SIZE) * BYTES_PER_SECTOR < $4000
section "Volume_ID_Root_Directory_Entry", rom0[(RESERVED_SECTOR_COUNT + FAT_SIZE) * BYTES_PER_SECTOR]
else
section "Volume_ID_Root_Directory_Entry", romx[(RESERVED_SECTOR_COUNT + FAT_SIZE) * BYTES_PER_SECTOR], bank[1]
endc

/*
 * DIR_Name
 *
 * Size: 11 Bytes
 */
DIR_Name:
    db "{VOLUME_LABEL}"

/*
 * DIR_Attr
 *
 * Size: 1 Byte
 */
DIR_Attr:
    db $08 ; ATTR_VOLUME_ID

/*
 * DIR_NTRes
 * Reserved. Must be set to 0.
 *
 * Size: 1 Byte
 */
DIR_NTRes:
    db $00

/*
 * DIR_CrtTimeTenth
 * (Optional)
 * Component of the file creation time. Count of
 * tenths of a second. Valid range is:
 * 0 <= DIR_CrtTimeTenth <= 199
 *
 * Size: 1 Byte
 */
DIR_CrtTimeTenth:
    db $00

/*
 * DIR_CrtTime
 * (Optional)
 * Creation time. Granularity is 2 seconds.
 *
 * Size: 2 Bytes
 */
DIR_CrtTime:
    dw $0000

/*
 * DIR_CrtDate
 * (Optional)
 * Creation date.
 *
 * Size: 2 Bytes
 */
DIR_CrtDate:
    dw $0000

/*
 * DIR_LstAccDate
 * (Optional)
 * Last access date. Last access is defined as a
 * read or write operation performed on the
 * file/directory described by this entry.
 *
 * Size: 2 Bytes
 */
DIR_LstAccDate:
    dw $0000

/*
 * DIR_FstClusHI
 * High word of first data cluster number for
 * file/directory described by this entry.
 * Only valid for volumes formatted FAT32. Must be
 * set to 0 on volumes formatted FAT12/FAT16.
 *
 * Size: 2 Bytes
 */
DIR_FstClusHI:
    dw $0000

/*
 * DIR_WrtTime
 * Last modification (write) time. Value must be
 * equal to DIR_CrtTime at file creation.
 *
 * Size: 2 Bytes
 */
DIR_WrtTime:
    dw (__UTC_HOUR__ << 11) + \
        (__UTC_MINUTE__ << 5) + \
        (__UTC_SECOND__ / 2)

/*
 * DIR_WrtDate
 * Last modification (write) date. Value must be
 * equal to DIR_CrtDate at file creation.
 *
 * Size: 2 Bytes
 */
DIR_WrtDate:
    dw (((__UTC_YEAR__ - 1980) & %011111111) << 9) + \
        (__UTC_MONTH__ << 5) + \
        __UTC_DAY__

/*
 * DIR_FstClusLO
 * Low word of first data cluster number for
 * file/directory described by this entry.
 * Zero in entry created by mkdosfs.
 *
 * Size: 2 Bytes
 */
DIR_FstClusLO:
    dw $0000

/*
 * DIR_FileSize
 * 32-bit quantity containing size in bytes of
 * file/directory described by this entry.
 * Zero in entry created by mkdosfs.
 *
 * Size: 4 Bytes
 */
DIR_FileSize:
    dl 0


/*
 * Game Boy Header specification found at https://gbdev.io/pandocs/The_Cartridge_Header.html
 */

section "Entry_point", rom0[$0100]
/*
 * After displaying the Nintendo logo, the built-in boot ROM jumps to
 * the address $0100, which should then jump to the actual main
 * program in the cartridge. Most commercial games fill this 4-byte
 * area with a nop instruction followed by a jp $0150.
 */
Entry_point:
    di
    jp Start

section "Nintendo_logo", rom0[$0104]
/*
 * This area contains a bitmap image that is displayed when the Game
 * Boy is powered on.
 * The Game Boy’s boot procedure first displays the logo and then
 * checks that it matches. If it doesn’t, the boot ROM locks itself
 * up.
 * The CGB and later models only check the top half of the logo (the
 * first $18 bytes).
 */
Nintendo_logo:
    db $CE, $ED, $66, $66, $CC, $0D, $00, $0B
    db $03, $73, $00, $83, $00, $0C, $00, $0D
    db $00, $08, $11, $1F, $88, $89, $00, $0E
    db $DC, $CC, $6E, $E6, $DD, $DD, $D9, $99
    db $BB, $BB, $67, $63, $6E, $0E, $EC, $CC
    db $DD, $DC, $99, $9F, $BB, $B9, $33, $3E

section "Title", rom0[$0134]
/*
 * These bytes contain the title of the game in upper case ASCII. If
 * the title is less than 16 characters long, the remaining bytes
 * should be padded with $00s.
 * Parts of this area actually have a different meaning on later
 * cartridges, reducing the actual title size to 15 ($0134–$0142) or
 * 11 ($0134–$013E) characters; see below.
 */
Title:
    db "{GBDOS_NAME_STRING}"
    ds $013F - @, 0

section "Manufacturer_code", rom0[$013F]
/*
 * In older cartridges these bytes were part of the Title (see
 * above). In newer cartridges they contain a 4-character
 * manufacturer code (in uppercase ASCII). The purpose of the
 * manufacturer code is unknown.
 */
Manufacturer_code:
    db "BMOR"

section "CGB_flag", rom0[$0143]
/*
 * In older cartridges this byte was part of the Title (see above).
 * The CGB and later models interpret this byte to decide whether to
 * enable Color mode (“CGB Mode”) or to fall back to monochrome
 * compatibility mode (“Non-CGB Mode”).
 */
CGB_flag:
    db $80

section "New_licensee_code", rom0[$0144]
/*
 * This area contains a two-character ASCII “licensee code”
 * indicating the game’s publisher. It is only meaningful if the Old
 * licensee is exactly $33 (which is the case for essentially all
 * games made after the SGB was released); otherwise, the old code
 * must be considered.
 */
New_licensee_code:
    db "00"

section "SGB_flag", rom0[$0146]
/*
 * This byte specifies whether the game supports SGB functions. The
 * SGB will ignore any command packets if this byte is set to a value
 * other than $03 (typically $00).
 */
SGB_flag:
    db $03

section "Cartridge_type", rom0[$0147]
/*
 * This byte indicates what kind of hardware is present on the
 * cartridge — most notably its mapper.
 */
Cartridge_type:
    db CARTRIDGE_TYPE

section "ROM_size", rom0[$0148]
/*
 * This byte indicates how much ROM is present on the cartridge. In
 * most cases, the ROM size is given by 32 KiB × (1 << <value>):
 */
ROM_size:
    db ROM_SIZE

section "RAM_size", rom0[$0149]
/*
 * This byte indicates how much RAM is present on the cartridge, if
 * any.
 * If the cartridge type does not include “RAM” in its name, this
 * should be set to 0. This includes MBC2, since its 512 × 4 bits of
 * memory are built directly into the mapper.
 */
RAM_size:
    db $00

section "Destination_code", rom0[$014A]
/*
 * This byte specifies whether this version of the game is intended
 * to be sold in Japan or elsewhere.
 */
Destination_code:
    db $00

section "Old_licensee_code", rom0[$014B]
/*
 * This byte is used in older (pre-SGB) cartridges to specify the
 * game’s publisher. However, the value $33 indicates that the New
 * licensee codes must be considered instead. (The SGB will ignore
 * any command packets unless this value is $33.)
 */
Old_licensee_code:
    db $33

section "Mask_ROM_version_number", rom0[$014C]
/*
 * This byte specifies the version number of the game. It is usually
 * $00.
 */
Mask_ROM_version_number:
    db GBDOS_MASK_ROM

section "Header_checksum", rom0[$014D]
/*
 * This byte contains an 8-bit checksum computed from the cartridge
 * header bytes $0134–014C.
 * The boot ROM verifies this checksum. If the byte at $014D does not
 * match the lower 8 bits of checksum, the boot ROM will lock up and
 * the program in the cartridge won’t run.
 */
Header_checksum:
    db $00 ; fixed by RGBFIX

section "Global_checksum", rom0[$014E]
/*
 * These bytes contain a 16-bit (big-endian) checksum simply computed
 * as the sum of all the bytes of the cartridge ROM (except these two
 * checksum bytes).
 * This checksum is not verified, except by Pokémon Stadium’s “GB
 * Tower” emulator (presumably to detect Transfer Pak errors).
 */
Global_checksum:
    dw $0000 ; fixed by RGBFIX

section "int_vblank", rom0[$0040]
int_vblank:
    push de
    ld de, r_inth_vblank
    jr common_interrupt

section "int_stat", rom0[$0048]
int_stat:
    push de
    ld de, r_inth_stat
    jr common_interrupt

section "int_timer", rom0[$0050]
int_timer:
    push de
    ld de, r_inth_timer
    jr common_interrupt

section "int_serial", rom0[$0058]
int_serial:
    push de
    ld de, r_inth_serial
    jr common_interrupt

section "int_joypad", rom0[$0060]
int_joypad:
    push de
    ld de, r_inth_joypad
common_interrupt:
    push af
    push bc
    push hl
    ld a, [de]
    ld l, a
    inc de
    ld a, [de]
    ld h, a
    ld bc, interrupt_return
    push bc
    jp hl

section "interrupt_return", rom0
interrupt_return:
    pop hl
    pop bc
    pop af
    pop de
    reti

section "default_interrupt_handler", rom0
default_interrupt_handler:
    ret

section "kernel_filename", rom0
kernel_filename:
    db "KERNEL  BIN"
end_kernel_filename:

section "Start", rom0
Start:
    ; Save all of the registers for future system identification
    ld sp, r_init_hl + 2
    push hl
    push de
    push bc
    push af

    ; Set all of the interrupt handlers to default
    ld hl, default_interrupt_handler
    push hl
    push hl
    push hl
    push hl
    push hl
    ld sp, $FFFE
    ei

    ld bc, ROOT_ENTRY_COUNT

next_entry:
    ld hl, - ROOT_ENTRY_COUNT
    add hl, bc
    ld a, h ; \
    cpl     ; |
    ld h, a ; |_ negate hl
    ld a, l ; |
    cpl     ; |
    ld l, a ; |
    inc hl  ; /
    push bc
    ld b, h
    ld c, l
    ld hl, 0
    ld de, 32
.mul_loop:
    ld a, b
    or c
    jr z, .mul_continue
    add hl, de
    dec bc
    jr .mul_loop
.mul_continue:
    ld de, (BYTES_PER_SECTOR * SECTORS_PER_CLUSTER) * ((NUMBER_OF_FATS * FAT_SIZE) + RESERVED_SECTOR_COUNT)
    add hl, de

    ld b, 11
    ld de, kernel_filename

.strcmp_loop
    ld a, b
    or a
    jp z, found_entry_1
    dec b
    ld a, [de]
    inc de
    ld c, a
    ldi a, [hl]
    cp c
    jr z, .strcmp_loop

    pop bc
    dec bc
    ld a, b
    or c
    jr nz, next_entry
    jp error ; Kernel not found

section "found_entry_1", rom0 ; This is split to fit better into the boot sector.
found_entry_1:
    pop bc; discard

    ; put cluster id in DE
    ld bc, 15
    add hl, bc
    ldi a, [hl]
    jp found_entry_2

section"found_entry_2", rom0
found_entry_2:
    ld e, a
    ldi a, [hl]
    ld d, a

    ; put file length in BC
    ldi a, [hl]
    ld c, a
    ldi a, [hl]
    ld b, a
    ldi a, [hl]
    or a
    jp nz, error ; Kernel too large
    ldi a, [hl]
    or a
    jp nz, error ; Kernel too large
    ld hl, -$1000
    add hl, bc
    jp c, error ; Kernel too large

    ; convert file length to sector count
    ld a, b
    and a, %1
    or c
    jr z, .round_length
    ld a, %10
    add a, b
    ld b, a
.round_length:
    srl b
    ld c, b
    xor a
    ld b, a

    ; set copy destination
    ld [r_copy_dest], a
    ld a, $C0
    ld [r_copy_dest + 1], a

    ; jump to kernel if no more to copy
copy_sector:
    ld a, c
    or a
    jp z, $C000

    ; Convert cluster id to rom bank number and switch banks
    ld hl, 2
    add hl, de
    ld a, 5 ; looping is more compact than unrolling here
.shift_right:
    dec a
    srl h ; 1
    rr l
    or a
    jr nz, .shift_right
    ld a, h
    ld [$3000], a ; MBC3 bank number gets overwritten with the second
    ld a, l       ; write at $2000. Without an MBC, these writes do
    ld [$2000], a ; nothing.

    ; Get top 2 bit of address
    or h
    jr z, .address ; A already zeroed out
    ld a, %01000000

    ; Convert cluster id to memory address
.address:
    ld hl, RESERVED_SECTOR_COUNT + FAT_SIZE
    add hl, de
    sla l
    push af
    ld a, l
    and %00111111 ; discard top 2 bits
    ld h, a
    pop af
    or h ; insert top 2 bits from before
    ld h, a
    xor a
    ld l, a

    ; copy cluster from ROM to WRAM
    push bc
    push de
    ld bc, BYTES_PER_SECTOR * SECTORS_PER_CLUSTER
    ld a, [r_copy_dest]
    ld e, a
    ld a, [r_copy_dest + 1]
    ld d, a
.loop:
    ld a, b
    or c
    jr z, .continue
    ldi a, [hl]
    ld [de], a
    inc de
    dec bc
    jr .loop
.continue:
    ld a, e
    ld [r_copy_dest], a
    ld a, d
    ld [r_copy_dest + 1], a
    pop de
    pop bc

    dec bc ; one less cluster to copy

get_next_cluster:
if FAT_TYPE == 12
    ld h, d
    ld l, e
    srl h ; divide by 2
    rr l
    add hl, de ; hl = de * 1.5
    bit 0, e ; save 0 bit for later in Z flag
    ld de, (RESERVED_SECTOR_COUNT * BYTES_PER_SECTOR)
    add hl, de ; add offset
    ldi a, [hl] ; same for both offsets
    jr nz, .odd_offset
    ld e, a
    ld a, [hl]
    and %00001111
    ld d, a
    jr .got_next_cluster
.odd_offset:
    swap a
    and %00001111
    ld e, a
    ld a, [hl]
    swap a
    and %11110000
    or e
    ld e, a
    ld a, [hl]
    swap a
    and %00001111
    ld d, a
.got_next_cluster:
else; FAT16
    sla e
    rl d
    ld hl, (RESERVED_SECTOR_COUNT * BYTES_PER_SECTOR)
    add hl, de
    ldi a, [hl]
    ld e, a
    ld a, [hl]
    ld e, a
endc
    jr copy_sector

section "error", rom0
error:
    ei
    nop
    halt
    nop
    jr error


section "Bootloader_registers", hram[$FF80]
r_inth_vblank:
    dw
r_inth_stat:
    dw
r_inth_timer:
    dw
r_inth_serial:
    dw
r_inth_joypad:
    dw
r_init_af:
    dw
r_init_bc:
    dw
r_init_de:
    dw
r_init_hl:
    dw
r_copy_dest:
    dw

/*
 * VHD Footer specifications found at https://github.com/libyal/libvhdi/blob/main/documentation/Virtual%20Hard%20Disk%20(VHD)%20image%20format.asciidoc
 * and https://pkg.go.dev/github.com/Microsoft/azure-vhd-utils/vhdcore/footer
 */

section "VHD_Signature", romx[VHD_ADDRESS + 0], bank[VHD_BANK]
/*
 * Signature (Cookie)
 *
 * Offset:  0
 * Size:    8
 */
VHD_Signature:
    db "conectix"

section "VHD_Features", romx[VHD_ADDRESS + 8], bank[VHD_BANK]
/*
 * Features
 * Value      | Description
 * -----------+--------------------
 * 0x00000000 | No features enabled
 * 0x00000001 | Temporary disk
 * 0x00000002 | Reserved
 *
 * Offset:  8
 * Size:    4
 */
VHD_Features:
    dl $00000000

section "VHD_Format_Version", romx[VHD_ADDRESS + 12], bank[VHD_BANK]
/*
 * Format Version
 * Where the upper 16-bit are the major version and the lower 16-bit the minor version
 *
 * Offset:  12
 * Size:    4
 */
VHD_Format_Version:
    dl $00010000

section "VHD_Next_Offset", romx[VHD_ADDRESS + 16], bank[VHD_BANK]
/*
 * Next Offset
 * In fixed disk images it should be set to 0xffffffffffffffff (-1).
 *
 * Offset:  16
 * Size:    8
 */
VHD_Next_Offset:
    db $FF, $FF, $FF, $FF
    db $FF, $FF, $FF, $FF

section "VHD_Modification_Time", romx[VHD_ADDRESS + 24], bank[VHD_BANK]
/*
 * Modificattion Time
 * Contains the number of seconds since January 1, 2000 00:00:00 UTC
 *
 * Offset:  24
 * Size:    4
 */
VHD_Modification_Time:
    ; This is a rather rough estimate...
    dl ((__UTC_YEAR__ - 2000) * 31556926)\ ; average year length
       + (__UTC_MONTH__ * 2629743)\ ; average month length
       + (__UTC_DAY__ * 86400)\
       + (__UTC_HOUR__ * 3600)\
       + (__UTC_MINUTE__ * 60)\
       + __UTC_SECOND__

section "VHD_Creator_Application", romx[VHD_ADDRESS + 28], bank[VHD_BANK]
/*
 * Creator Application
 * Known values:
 * Value   | Description
 * --------+--------------------------
 * "d2v",0 | Disk2vhd
 * "qemu"  | Qemu
 * "vpc "  | Virtual PC
 * "vs  "  | Virtual Server
 * "win "  | Windows (Disk Management)
 *
 * Offset:  28
 * Size:    4
 */
VHD_Creator_Application:
    db strsub("{GBDOS_NAME_STRING}    ", 0, 4) ; pad with spaces before truncating

section "VHD_Creator_Version", romx[VHD_ADDRESS + 32], bank[VHD_BANK]
/*
 * Creator Version
 * Where the upper 16-bit are the major version and the lower 16-bit the minor version
 *
 * Offset:  32
 * Size:    4
 */
VHD_Creator_Version:
    dw GBDOS_MAJOR_RELEASE_NUMBER
    dw GBDOS_MINOR_RELEASE_NUMBER

section "VHD_Creator_Operating_System", romx[VHD_ADDRESS + 36], bank[VHD_BANK]
/*
 * Creator (host) operating system
 * Two known values are "Mac " and "Wi2k"
 *
 * Offset:  36
 * Size:    4
 */
VHD_Creator_Operating_System:
    db strsub("UNIX    ", 0, 4)

section "VHD_Disk_Size", romx[VHD_ADDRESS + 40], bank[VHD_BANK]
/*
 * Disk Size
 * Physical Size
 * Contains the size of the disk in bytes
 *
 * Offset:  40
 * Size:    8
 */
VHD_Disk_Size:
    dl 0, IMAGE_SIZE

section "VHD_Data_Size", romx[VHD_ADDRESS + 48], bank[VHD_BANK]
/*
 * Data Size
 * Virtual Size
 * Contains the size of the data in bytes
 *
 * Offset:  48
 * Size:    8
 */
VHD_Data_Size:
    dl 0, IMAGE_SIZE

section "VHD_Disk_Geometry", romx[VHD_ADDRESS + 56], bank[VHD_BANK]
/*
 * Disk Geometry
 * The disk geometry is 4 bytes of size and consists of:
 * Offset | Size | Description
 * -------+------+------------
 * 0      | 2    | Number of cylinders
 * 2      | 1    | Number of heads
 * 3      | 1    | Number of sectors per track (cylinder)
 *
 * Offset:  56
 * Size:    4
 */
VHD_Disk_Geometry:
    dw ROM_BANK_COUNT
    db 1
    db $4000 / BYTES_PER_SECTOR; Interpreted as sectors per bank

section "VHD_Disk_Type", romx[VHD_ADDRESS + 60], bank[VHD_BANK]
/*
 * Disk Type
 * Value | Description
 * ------+------------
 * 0     | None
 * 1     | Reserved (deprecated)
 * 2     | Fixed hard disk
 * 3     | Dynamic hard disk
 * 4     | Differential hard disk
 * 5     | Reserved (deprecated)
 * 6     | Reserved (deprecated)
 *
 * Offset: 60
 * Size: 4
 */
VHD_Disk_Type:
    dl 2

section "VHD_Checksum", romx[VHD_ADDRESS + 64], bank[VHD_BANK]
/*
 * Checksum
 * It is just a one’s complement of the sum of all the bytes in the
 * footer without the checksum field.
 *
 * Offset:  64
 * Size:    4
 */
VHD_Checksum:
    ; Currently we don't have a way to calculate this.
    ; Maybe it can be done with a untility program similar to RGBFIX?
    dl 0

section "VHD_Identifier", romx[VHD_ADDRESS + 68], bank[VHD_BANK]
/*
 * Identifier
 * Contains a big-endian GUID
 *
 * Offset:  68
 * Size:    16
 */
VHD_Identifier:
    ; Currently we don't have a good way to make a GUID.
    ; Here we're making a value up from various sources.
    dw __UTC_YEAR__ ; 2B, 2B
    db __UTC_MONTH__; 1B, 3B
    db __UTC_DAY__  ; 1B, 4B
    db __UTC_HOUR__ ; 1B, 5B
    db __UTC_MINUTE__   ; 1B, 6B
    db __UTC_SECOND__   ; 1B, 7B
    db __RGBDS_MAJOR__  ; 1B, 8B
    db __RGBDS_MINOR__  ; 1B, 9B
    db __RGBDS_PATCH__  ; 1B, 10B
    db GBDOS_MAJOR_RELEASE_NUMBER   ; 1B, 11B
    db GBDOS_MINOR_RELEASE_NUMBER   ; 1B, 12B
    db strsub("{GBDOS_NAME_STRING}    ", 0, 4)  ; 4B, 16B

section "VHD_Saved_State", romx[VHD_ADDRESS + 84], bank[VHD_BANK]
/*
 * Saved State
 * Flag to indicate the image is in saved state.
 *
 * Offset:  84
 * Size:    1
 */
VHD_Saved_State:
    ; unknown what value this should be
    db 0

section "VHD_Reserved", romx[VHD_ADDRESS + 85], bank[VHD_BANK]
/*
 * Reserved
 * (empty values)
 *
 * Offset:  85
 * Size:    427
 */
VHD_Reserved:
    ds 427, $00
